# coding: utf-8
# Author: Felix Fontein <felix@fontein.de>
# License: GPLv3+
# Copyright: Ansible Project, 2020

"""
Entrypoint to the antsibull-changelog script.
"""

import argparse
import os.path
import sys
import traceback

from typing import Any, List

try:
    import argcomplete
    HAS_ARGCOMPLETE = True
except ImportError:
    HAS_ARGCOMPLETE = False

from antsibull_changelog.lint import lint_changelog_yaml
from antsibull_changelog.logger import setup_logger

from antsibull_core.args import get_toplevel_parser, normalize_toplevel_options

from antsibull_docs.lint_extra_docs import lint_collection_extra_docs_files
from antsibull_docs.collection_links import lint_collection_links


def run(args: List[str]) -> int:
    """
    Main program entry point.
    """
    print(
        'WARNING: `antsibull-lint` is deprecated. Use `antsibull-changelog lint-changelog-yaml`'
        ' or `antsibull-docs lint-collection-docs` depending on your use-case.',
        file=sys.stderr)
    verbosity = 0
    try:
        program_name = os.path.basename(args[0])
        parser = get_toplevel_parser(
            prog=program_name,
            package='antsibull',
            description='Linting tool')

        common = argparse.ArgumentParser(add_help=False)
        common.add_argument('-v', '--verbose',
                            action='count',
                            default=0,
                            help='increase verbosity of output')

        subparsers = parser.add_subparsers(dest='command')
        # Make sure that our code below is used instead of the default code. Our code uses
        # parser.print_help(), which shows the deprecation messages.
        # subparsers.required = True

        changelog_yaml = subparsers.add_parser('changelog-yaml',
                                               parents=[common],
                                               help='changelogs/changelog.yaml linter. WARNING:'
                                                    ' this is deprecated; use `antsibull-changelog'
                                                    ' lint-changelog-yaml` instead!')
        changelog_yaml.set_defaults(func=command_lint_changelog)

        changelog_yaml.add_argument('changelog_yaml_path',
                                    metavar='/path/to/changelog.yaml',
                                    help='path to changelogs/changelog.yaml')

        changelog_yaml.add_argument('--no-semantic-versioning', action='store_true',
                                    help='Assume that use_semantic_versioning=false in the'
                                         ' changelog config. Do not use this for Ansible'
                                         ' collections!')

        collection_docs = subparsers.add_parser('collection-docs',
                                                parents=[common],
                                                help='Collection extra docs linter for inclusion'
                                                     ' in docsite. WARNING: this is deprecated;'
                                                     ' use `antsibull-docs lint-collection-docs`'
                                                     ' instead!')
        collection_docs.set_defaults(func=command_lint_collection_docs)

        collection_docs.add_argument('collection_root_path',
                                     metavar='/path/to/collection',
                                     help='path to collection (directory that includes galaxy.yml)')

        if HAS_ARGCOMPLETE:
            argcomplete.autocomplete(parser)

        arguments = parser.parse_args(args[1:])

        if getattr(arguments, 'func', None) is None:
            parser.print_help()
            return 2

        normalize_toplevel_options(arguments)

        verbosity = arguments.verbose
        setup_logger(verbosity)

        return arguments.func(arguments)
    except SystemExit as e:
        return e.code
    except Exception:  # pylint: disable=broad-except
        if verbosity > 0:
            traceback.print_exc()
        else:
            print('ERROR: Uncaught exception. Run with -v to see traceback.')
        return 1


def command_lint_changelog(args: Any) -> int:
    """
    Validate a changelogs/changelog.yaml file.

    :arg args: Parsed arguments
    """
    print(
        'WARNING: `antsibull-lint changelog-yaml` is deprecated.'
        ' Use `antsibull-changelog lint-changelog-yaml` instead.',
        file=sys.stderr)

    errors = lint_changelog_yaml(
        args.changelog_yaml_path, no_semantic_versioning=args.no_semantic_versioning)

    messages = sorted(set(f'{error[0]}:{error[1]}:{error[2]}: {error[3]}' for error in errors))

    for message in messages:
        print(message)

    return 3 if messages else 0


def command_lint_collection_docs(args: Any) -> int:
    """
    Validate docs/docsite/rst/ in a collection.

    :arg args: Parsed arguments
    """
    print(
        'WARNING: `antsibull-lint collection-docs` is deprecated.'
        ' Use `antsibull-docs lint-collection-docs` instead.',
        file=sys.stderr)

    errors = lint_collection_extra_docs_files(args.collection_root_path)
    errors.extend(lint_collection_links(args.collection_root_path))

    messages = sorted(set(f'{error[0]}:{error[1]}:{error[2]}: {error[3]}' for error in errors))

    for message in messages:
        print(message)

    return 3 if messages else 0


def main() -> int:
    """
    Entrypoint called from the script.

    console_scripts call functions which take no parameters.  However, it's hard to test a function
    which takes no parameters so this function lightly wraps :func:`run`, which actually does the
    heavy lifting.

    :returns: A program return code.

    Return codes:
        :0: Success
        :1: Unhandled error.  See the Traceback for more information.
        :2: There was a problem with the command line arguments
        :3: Linting failed
        :4: Needs to be run on a newer version of Python
    """
    if sys.version_info < (3, 6):
        print('Needs Python 3.6 or later')
        return 4

    return run(sys.argv)


if __name__ == '__main__':
    sys.exit(main())



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Build Ansible packages."""

import asyncio
import datetime
import os
import os.path
import shutil
import tempfile
import typing as t
from functools import partial

import aiofiles
import aiohttp
import asyncio_pool
import sh
from jinja2 import Template
from packaging.version import Version as PypiVer
from semantic_version import Version as SemVer, SimpleSpec as SemVerSpec

from antsibull_core import app_context
from antsibull_core.ansible_core import get_ansible_core_package_name
from antsibull_core.collections import install_separately, install_together
from antsibull_core.dependency_files import BuildFile, DependencyFileData, DepsFile
from antsibull_core.galaxy import CollectionDownloader, GalaxyClient
from antsibull_core.utils.io import write_file
from antsibull_core.yaml import store_yaml_file

from .build_changelog import ReleaseNotes
from .changelog import ChangelogData, get_changelog
from .dep_closure import check_collection_dependencies
from .utils.get_pkg_data import get_antsibull_data


#
# Common code
#


async def get_collection_versions(deps: t.Mapping[str, str],
                                  galaxy_url: str,
                                  ) -> t.Dict[str, SemVer]:
    """
    Retrieve the latest version of each collection.

    :arg deps: Mapping of collection name to a version specification.
    :arg galaxy_url: The url for the galaxy server to use.
    :returns: Dict mapping collection name to latest version.
    """
    requestors = {}
    async with aiohttp.ClientSession() as aio_session:
        lib_ctx = app_context.lib_ctx.get()
        async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
            client = GalaxyClient(aio_session, galaxy_server=galaxy_url)
            for collection_name, version_spec in deps.items():
                requestors[collection_name] = await pool.spawn(
                    client.get_latest_matching_version(collection_name, version_spec, pre=True))

            responses = await asyncio.gather(*requestors.values())

    # Note: Python dicts have a stable sort order and since we haven't modified the dict since we
    # used requestors.values() to generate responses, requestors and responses therefore have
    # a matching order.
    included_versions: t.Dict[str, SemVer] = {}
    for collection_name, version in zip(requestors, responses):
        included_versions[collection_name] = version

    return included_versions


async def download_collections(versions: t.Mapping[str, SemVer],
                               galaxy_url: str,
                               download_dir: str,
                               collection_cache: t.Optional[str] = None,
                               ) -> None:
    requestors = {}
    async with aiohttp.ClientSession() as aio_session:
        lib_ctx = app_context.lib_ctx.get()
        async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
            downloader = CollectionDownloader(aio_session, download_dir,
                                              collection_cache=collection_cache,
                                              galaxy_server=galaxy_url)
            for collection_name, version in versions.items():
                requestors[collection_name] = await pool.spawn(
                    downloader.download(collection_name, version))

            await asyncio.gather(*requestors.values())


#
# Single sdist for ansible
#

def copy_boilerplate_files(package_dir: str) -> None:
    gpl_license = get_antsibull_data('gplv3.txt')
    with open(os.path.join(package_dir, 'COPYING'), 'wb') as f:
        f.write(gpl_license)

    readme = get_antsibull_data('ansible-readme.rst')
    with open(os.path.join(package_dir, 'README.rst'), 'wb') as f:
        f.write(readme)


def write_manifest(package_dir: str,
                   release_notes: t.Optional[ReleaseNotes] = None,
                   debian: bool = False) -> None:
    manifest_file = os.path.join(package_dir, 'MANIFEST.in')
    with open(manifest_file, 'w', encoding='utf-8') as f:
        f.write('include COPYING\n')
        f.write('include README\n')
        f.write('include build-ansible.sh\n')
        if release_notes:
            f.write(f'include {release_notes.changelog_filename}\n')
            f.write(f'include {release_notes.porting_guide_filename}\n')
        if debian:
            f.write('include debian/*\n')
        f.write('recursive-include ansible_collections/ **\n')


def write_release_py(ansible_version: PypiVer, ansible_collections_dir: str) -> None:
    release_filename = os.path.join(ansible_collections_dir, 'ansible_release.py')

    release_tmpl = Template(get_antsibull_data('ansible-release_py.j2').decode('utf-8'))
    release_contents = release_tmpl.render(version=ansible_version)

    with open(release_filename, 'w', encoding='utf-8') as f:
        f.write(release_contents)


def write_ansible_community_py(ansible_version: PypiVer, ansible_collections_dir: str) -> None:
    release_filename = os.path.join(ansible_collections_dir, 'ansible_community.py')

    release_tmpl = Template(get_antsibull_data('ansible-community.py.j2').decode('utf-8'))
    release_contents = release_tmpl.render(version=ansible_version)

    with open(release_filename, 'w', encoding='utf-8') as f:
        f.write(release_contents + '\n')


def write_setup(ansible_version: PypiVer,
                ansible_core_version: PypiVer,
                collection_exclude_paths: t.List[str],
                collection_deps: str,
                package_dir: str) -> None:
    setup_filename = os.path.join(package_dir, 'setup.py')

    setup_tmpl = Template(get_antsibull_data('ansible-setup_py.j2').decode('utf-8'))
    setup_contents = setup_tmpl.render(
        version=ansible_version,
        ansible_core_package_name=get_ansible_core_package_name(ansible_core_version),
        ansible_core_version=ansible_core_version,
        collection_exclude_paths=collection_exclude_paths,
        collection_deps=collection_deps,
        PypiVer=PypiVer,
    )

    with open(setup_filename, 'w', encoding='utf-8') as f:
        f.write(setup_contents)


def write_python_build_files(ansible_version: PypiVer,
                             ansible_core_version: PypiVer,
                             collection_exclude_paths: t.List[str],
                             collection_deps: str,
                             package_dir: str,
                             release_notes: t.Optional[ReleaseNotes] = None,
                             debian: bool = False) -> None:
    copy_boilerplate_files(package_dir)
    write_manifest(package_dir, release_notes, debian)
    write_setup(
        ansible_version, ansible_core_version, collection_exclude_paths, collection_deps,
        package_dir)


def write_debian_directory(ansible_version: PypiVer,
                           ansible_core_version: PypiVer,
                           package_dir: str) -> None:
    debian_dir = os.path.join(package_dir, 'debian')
    os.mkdir(debian_dir, mode=0o700)
    debian_files = ('changelog.j2', 'control.j2', 'copyright', 'rules')
    ansible_core_package_name = get_ansible_core_package_name(ansible_core_version)
    for filename in debian_files:
        # Don't use os.path.join here, the get_data docs say it should be
        # slash-separated.
        src_pkgfile = 'debian/' + filename
        data = get_antsibull_data(src_pkgfile).decode('utf-8')

        if filename.endswith('.j2'):
            filename = filename.replace('.j2', '')
            # If the file is a template, send it in vars it might need
            # and update 'data' to be the result.
            tmpl = Template(data)
            data = tmpl.render(
                version=str(ansible_version),
                date=datetime.datetime.utcnow().strftime('%a, %d %b %Y %T +0000'),
                ansible_core_package_name=ansible_core_package_name,
            )

        with open(os.path.join(debian_dir, filename), 'w', encoding='utf-8') as f:
            f.write(data)


def write_galaxy_requirements(filename: str, included_versions: t.Mapping[str, str]) -> None:
    galaxy_reqs = []
    for collection, version in sorted(included_versions.items()):
        galaxy_reqs.append({
            'name': collection,
            'version': version,
            'source': 'https://galaxy.ansible.com'
        })

    store_yaml_file(filename, {
        'collections': galaxy_reqs,
    })


def make_dist(ansible_dir: str, dest_dir: str) -> None:
    # pyre-ignore[16]
    sh.python('setup.py', 'sdist', _cwd=ansible_dir)  # pylint:disable=no-member
    dist_dir = os.path.join(ansible_dir, 'dist')
    files = os.listdir(dist_dir)
    if len(files) != 1:
        raise Exception('python setup.py sdist should only have created one file')

    shutil.move(os.path.join(dist_dir, files[0]), dest_dir)


def make_dist_with_wheels(ansible_dir: str, dest_dir: str) -> None:
    # pyre-ignore[16]
    sh.python('setup.py', 'sdist', 'bdist_wheel', _cwd=ansible_dir)  # pylint:disable=no-member
    dist_dir = os.path.join(ansible_dir, 'dist')
    files = os.listdir(dist_dir)
    tarball_count = 0
    wheel_count = 0
    for file in files:
        if file.endswith('.tar') or file.endswith('.tar.gz'):
            tarball_count += 1
        elif file.endswith('.whl'):
            wheel_count += 1
        else:
            tarball_count = 2  # the number is wrong, but this triggers an error
            break
    if tarball_count != 1 or wheel_count == 0:
        raise Exception(
            "python setup.py sdist bdist_wheel should have created exactly one tarball and at"
            f" least one wheel (got {files})")

    for file in files:
        shutil.move(os.path.join(dist_dir, file), dest_dir)


def write_build_script(ansible_version: PypiVer,
                       ansible_core_version: PypiVer,
                       package_dir: str) -> None:
    """Write a build-script that tells how to build this tarball."""
    build_ansible_filename = os.path.join(package_dir, 'build-ansible.sh')

    build_ansible_tmpl = Template(get_antsibull_data('build-ansible.sh.j2').decode('utf-8'))
    build_ansible_contents = build_ansible_tmpl.render(version=ansible_version,
                                                       ansible_core_version=ansible_core_version)

    with open(build_ansible_filename, 'w', encoding='utf-8') as f:
        f.write(build_ansible_contents)
    os.chmod(build_ansible_filename, mode=0o755)


def build_single_command() -> int:
    # This is deprecated; in the future users will have to first run prepare, and then
    # rebuild-single.
    result = prepare_command()
    if result != 0:
        return result

    return rebuild_single_command()


def prepare_command() -> int:
    app_ctx = app_context.app_ctx.get()

    build_filename = os.path.join(app_ctx.extra['data_dir'], app_ctx.extra['build_file'])
    build_file = BuildFile(build_filename)
    build_ansible_version, ansible_core_version, deps = build_file.parse()
    ansible_core_version = PypiVer(ansible_core_version)

    # If we're building a feature frozen release (betas and rcs) then we need to
    # change the upper version limit to not include new features.
    if app_ctx.extra['feature_frozen']:
        old_deps, deps = deps, {}
        # For each collection that's listed...
        for collection_name, spec in old_deps.items():
            spec = SemVerSpec(spec)
            new_clauses = []
            min_version = None

            # Look at each clause of the version specification
            for clause in spec.clause.clauses:
                if clause.operator in ('<', '<='):
                    # Omit the upper bound as we're replacing it
                    continue

                if clause.operator == '>=':
                    # Save the lower bound so we can write out a new compatible version
                    min_version = clause.target

                new_clauses.append(str(clause))

            if min_version is None:
                raise ValueError(f'No minimum version specified for {collection_name}: {spec}')

            new_clauses.append(f'<{min_version.major}.{min_version.minor + 1}.0')
            deps[collection_name] = ','.join(new_clauses)

    included_versions = asyncio.run(get_collection_versions(deps, app_ctx.galaxy_url))

    if not str(app_ctx.extra['ansible_version']).startswith(build_ansible_version):
        print(f'{build_filename} is for version {build_ansible_version} but we need'
              f' {app_ctx.extra["ansible_version"].major}'
              f'.{app_ctx.extra["ansible_version"].minor}')
        return 1

    dependency_data = DependencyFileData(
        str(app_ctx.extra['ansible_version']),
        str(ansible_core_version),
        {collection: str(version) for collection, version in included_versions.items()})

    # Get Ansible changelog, add new release
    ansible_changelog = ChangelogData.ansible(
        app_ctx.extra['data_dir'], app_ctx.extra['dest_data_dir'])
    date = datetime.date.today()
    ansible_changelog.add_ansible_release(
        str(app_ctx.extra['ansible_version']),
        date,
        f'Release Date: {date}'
        f'\n\n'
        f'`Porting Guide <https://docs.ansible.com/ansible/devel/porting_guides.html>`_')
    ansible_changelog.changes.save()

    # Write dependency file
    deps_filename = os.path.join(app_ctx.extra['dest_data_dir'], app_ctx.extra['deps_file'])
    deps_file = DepsFile(deps_filename)
    deps_file.write(
        dependency_data.ansible_version,
        dependency_data.ansible_core_version,
        dependency_data.deps)

    # Write Galaxy requirements.yml file
    galaxy_filename = os.path.join(app_ctx.extra['dest_data_dir'], app_ctx.extra['galaxy_file'])
    write_galaxy_requirements(galaxy_filename, dependency_data.deps)

    return 0


def compile_collection_exclude_paths(collection_names: t.Collection[str],
                                     collection_root: str) -> t.Tuple[t.List[str], t.List[str]]:
    result = set()
    ignored_files = set()
    all_files = []
    for collection_name in collection_names:
        namespace, name = collection_name.split('.', 1)
        prefix = f"{namespace}/{name}/"

        # Check files
        collection_dir = os.path.join(collection_root, namespace, name)
        all_files.clear()
        for directory, _, files in os.walk(collection_dir):
            directory = os.path.relpath(directory, collection_dir)
            for file in files:
                all_files.append(os.path.normpath(os.path.join(directory, file)))

        def ignore_file(prefix: str, filename: str):  # pylint: disable=unused-variable
            if filename in all_files:
                result.add(prefix + filename)
                ignored_files.add(prefix + filename)

        def ignore_start(prefix: str, start: str):
            matching_files = [file for file in all_files if file.startswith(start)]
            if matching_files:
                result.add(prefix + start + '*')
                ignored_files.update([prefix + file for file in matching_files])

        def ignore_directory(prefix: str, directory: str):
            directory = directory.rstrip('/') + '/'
            matching_files = [file for file in all_files if file.startswith(directory)]
            if matching_files:
                result.add(prefix + directory + '*')
                ignored_files.update([prefix + file for file in matching_files])

        ignore_start(prefix, '.')
        ignore_directory(prefix, 'docs')
        ignore_directory(prefix, 'tests')
    return sorted(result), sorted(ignored_files)


def rebuild_single_command() -> int:
    app_ctx = app_context.app_ctx.get()

    deps_filename = os.path.join(app_ctx.extra['data_dir'], app_ctx.extra['deps_file'])
    deps_file = DepsFile(deps_filename)
    dependency_data = deps_file.parse()

    # Determine included collection versions
    ansible_core_version = PypiVer(dependency_data.ansible_core_version)
    included_versions = {
        collection: SemVer(version)
        for collection, version in dependency_data.deps.items()
    }

    with tempfile.TemporaryDirectory() as tmp_dir:
        download_dir = os.path.join(tmp_dir, 'collections')
        os.mkdir(download_dir, mode=0o700)

        # Download included collections
        asyncio.run(download_collections(included_versions, app_ctx.galaxy_url,
                                         download_dir, app_ctx.collection_cache))

        # Get Ansible changelog, add new release
        ansible_changelog = ChangelogData.ansible(
            app_ctx.extra['data_dir'], app_ctx.extra['dest_data_dir'])

        # Get changelog and porting guide data
        changelog = get_changelog(
            app_ctx.extra['ansible_version'],
            deps_dir=app_ctx.extra['data_dir'],
            deps_data=[dependency_data],
            collection_cache=app_ctx.collection_cache,
            ansible_changelog=ansible_changelog)

        # Create package and collections directories
        package_dir = os.path.join(tmp_dir, f'ansible-{app_ctx.extra["ansible_version"]}')
        os.mkdir(package_dir, mode=0o700)
        ansible_collections_dir = os.path.join(package_dir, 'ansible_collections')
        os.mkdir(ansible_collections_dir, mode=0o700)

        # Write the ansible release info to the collections dir
        write_release_py(app_ctx.extra['ansible_version'], ansible_collections_dir)

        # Write the ansible-community CLI program (starting with Ansible 6.0.0rc1)
        if app_ctx.extra['ansible_version'] >= PypiVer('6.0.0rc1'):
            write_ansible_community_py(app_ctx.extra['ansible_version'], ansible_collections_dir)

        # Install collections
        # TODO: PY3.8:
        # collections_to_install = [p for f in os.listdir(download_dir)
        #                           if os.path.isfile(p := os.path.join(download_dir, f))]
        collections_to_install = []
        for collection in os.listdir(download_dir):
            path = os.path.join(download_dir, collection)
            if os.path.isfile(path):
                collections_to_install.append(path)

        asyncio.run(install_together(collections_to_install, ansible_collections_dir))

        # Compose and write release notes to destination directory
        release_notes = ReleaseNotes.build(changelog)
        release_notes.write_changelog_to(package_dir)
        release_notes.write_porting_guide_to(package_dir)

        # Write changelog and porting guide also to destination directory
        release_notes.write_changelog_to(app_ctx.extra['dest_data_dir'])
        release_notes.write_porting_guide_to(app_ctx.extra['dest_data_dir'])

        # pylint:disable-next=unused-variable
        collection_exclude_paths, collection_ignored_files = compile_collection_exclude_paths(
            dependency_data.deps, ansible_collections_dir)

        # TODO: do something with collection_ignored_files

        # Write build scripts and files
        write_build_script(app_ctx.extra['ansible_version'], ansible_core_version, package_dir)
        write_python_build_files(app_ctx.extra['ansible_version'], ansible_core_version,
                                 collection_exclude_paths, '', package_dir, release_notes,
                                 app_ctx.extra['debian'])
        if app_ctx.extra['debian']:
            write_debian_directory(app_ctx.extra['ansible_version'], ansible_core_version,
                                   package_dir)

        # Create source distribution
        if app_ctx.extra["ansible_version"].major < 6:
            make_dist(package_dir, app_ctx.extra['sdist_dir'])
        else:
            make_dist_with_wheels(package_dir, app_ctx.extra['sdist_dir'])

        # Check dependencies
        dep_errors = check_collection_dependencies(os.path.join(package_dir, 'ansible_collections'))

        if dep_errors:
            print('WARNING: found collection dependency errors!')
            for error in dep_errors:
                print(f'WARNING: {error}')

    return 0


#
# Code to make one sdist per collection
#


async def write_collection_readme(collection_name: str, package_dir: str) -> None:
    readme_tmpl = Template(get_antsibull_data('collection-readme.j2').decode('utf-8'))
    readme_contents = readme_tmpl.render(collection_name=collection_name)

    readme_filename = os.path.join(package_dir, 'README.rst')
    await write_file(readme_filename, readme_contents)


async def write_collection_setup(name: str, version: str, package_dir: str) -> None:
    setup_filename = os.path.join(package_dir, 'setup.py')

    setup_tmpl = Template(get_antsibull_data('collection-setup_py.j2').decode('utf-8'))
    setup_contents = setup_tmpl.render(version=version, name=name)

    await write_file(setup_filename, setup_contents)


async def write_collection_manifest(package_dir: str) -> None:
    manifest_file = os.path.join(package_dir, 'MANIFEST.in')
    async with aiofiles.open(manifest_file, 'w') as f:
        await f.write('include README.rst\n')
        await f.write('recursive-include ansible_collections/ **\n')


async def make_collection_dist(name: str,
                               version: str,
                               package_dir: str,
                               dest_dir: str) -> None:
    # Copy boilerplate into place
    await write_collection_readme(name, package_dir)
    await write_collection_setup(name, version, package_dir)
    await write_collection_manifest(package_dir)

    loop = asyncio.get_running_loop()

    # Create the python sdist
    # pyre-ignore[16], pylint:disable=no-member
    await loop.run_in_executor(None, partial(sh.python, 'setup.py', 'sdist', _cwd=package_dir))
    dist_dir = os.path.join(package_dir, 'dist')
    files = os.listdir(dist_dir)
    if len(files) != 1:
        raise Exception('python setup.py sdist should only have created one file')

    dist_file = os.path.join(dist_dir, files[0])
    await loop.run_in_executor(None, shutil.move, dist_file, dest_dir)


async def make_collection_dists(dest_dir: str, collection_dirs: t.List[str]) -> None:
    dist_creators = []
    lib_ctx = app_context.lib_ctx.get()
    async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
        for collection_dir in collection_dirs:
            dir_name_only = os.path.basename(collection_dir)
            dummy_, dummy_, name, version = dir_name_only.split('-', 3)

            dist_creators.append(await pool.spawn(
                make_collection_dist(name, version, collection_dir, dest_dir)))

        await asyncio.gather(*dist_creators)


def build_multiple_command() -> int:
    app_ctx = app_context.app_ctx.get()

    build_filename = os.path.join(app_ctx.extra['data_dir'], app_ctx.extra['build_file'])
    build_file = BuildFile(build_filename)
    build_ansible_version, ansible_core_version, deps = build_file.parse()
    ansible_core_version = PypiVer(ansible_core_version)

    # TODO: implement --feature-frozen support

    if not str(app_ctx.extra['ansible_version']).startswith(build_ansible_version):
        print(f'{build_filename} is for version {build_ansible_version} but we need'
              f' {app_ctx.extra["ansible_version"].major}'
              f'.{app_ctx.extra["ansible_version"].minor}')
        return 1

    with tempfile.TemporaryDirectory() as tmp_dir:
        download_dir = os.path.join(tmp_dir, 'collections')
        os.mkdir(download_dir, mode=0o700)

        included_versions = asyncio.run(get_collection_versions(deps, app_ctx.galaxy_url))
        asyncio.run(
            download_collections(included_versions, app_ctx.galaxy_url, download_dir,
                                 app_ctx.collection_cache))
        # TODO: PY3.8:
        # collections_to_install = [p for f in os.listdir(download_dir)
        #                           if os.path.isfile(p := os.path.join(download_dir, f))]
        collections_to_install = []
        for collection in os.listdir(download_dir):
            path = os.path.join(download_dir, collection)
            if os.path.isfile(path):
                collections_to_install.append(path)

        collection_dirs = asyncio.run(install_separately(collections_to_install, download_dir))
        asyncio.run(make_collection_dists(app_ctx.extra['sdist_dir'], collection_dirs))

        # Create the ansible package that deps on the collections we just wrote
        package_dir = os.path.join(tmp_dir, f'ansible-{app_ctx.extra["ansible_version"]}')
        os.mkdir(package_dir, mode=0o700)
        ansible_collections_dir = os.path.join(package_dir, 'ansible_collections')
        os.mkdir(ansible_collections_dir, mode=0o700)

        # Construct the list of dependent collection packages
        collection_deps = []
        for collection, version in sorted(included_versions.items()):
            collection_deps.append(f"        '{collection}>={version},<{version.next_major()}'")
        collection_deps = '\n' + ',\n'.join(collection_deps)
        write_build_script(app_ctx.extra['ansible_version'], ansible_core_version, package_dir)
        write_python_build_files(app_ctx.extra['ansible_version'], ansible_core_version,
                                 [], collection_deps, package_dir)

        make_dist(package_dir, app_ctx.extra['sdist_dir'])

    # Write the deps file
    deps_filename = os.path.join(app_ctx.extra['dest_data_dir'], app_ctx.extra['deps_file'])
    deps_file = DepsFile(deps_filename)
    deps_file.write(app_ctx.extra['ansible_version'], ansible_core_version, included_versions)

    return 0



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2021
"""Check collection dependencies."""

import json
import pathlib

from collections import namedtuple
from typing import Dict, List, Mapping

from semantic_version import Version as SemVer, SimpleSpec as SemVerSpec

from antsibull_core import app_context


CollectionRecord = namedtuple('CollectionRecord', ('version', 'dependencies'))


def parse_manifest(collection_dir: pathlib.Path) -> Mapping[str, CollectionRecord]:
    '''Parse MANIFEST.json for a collection.'''
    manifest = collection_dir.joinpath('MANIFEST.json')
    with manifest.open() as f:
        manifest_data = json.load(f)['collection_info']

    collection_record = {
        f'{manifest_data["namespace"]}.{manifest_data["name"]}':
            CollectionRecord(manifest_data['version'], manifest_data['dependencies'])
    }

    return collection_record


def analyze_deps(collections: Mapping[str, CollectionRecord]) -> List[str]:
    '''Analyze dependencies of a set of collections. Return list of errors found.'''
    errors = []

    # Look at dependencies
    # make sure their dependencies are found
    for collection_name, collection_info in collections.items():
        for dep_name, dep_version_spec in collection_info.dependencies.items():
            if dep_name not in collections:
                errors.append(f'{collection_name} missing: {dep_name} ({dep_version_spec})')
                continue

            dependency_version = SemVer(collections[dep_name].version)
            if dependency_version not in SemVerSpec(dep_version_spec):
                errors.append(f'{collection_name} version_conflict:'
                              f' {dep_name}-{str(dependency_version)} but needs'
                              f' {dep_version_spec}')
                continue

    return errors


def check_collection_dependencies(collection_root: str) -> List[str]:
    '''Analyze dependencies between collections in a collection root.'''
    ansible_collection_dir = pathlib.Path(collection_root)
    errors = []

    collections: Dict[str, CollectionRecord] = {}
    for namespace_dir in (n for n in ansible_collection_dir.iterdir() if n.is_dir()):
        for collection_dir in (c for c in namespace_dir.iterdir() if c.is_dir()):
            try:
                collections.update(parse_manifest(collection_dir))
            except FileNotFoundError:
                errors.append(f'{collection_dir} is not a valid collection')

    errors.extend(analyze_deps(collections))
    return errors


def validate_dependencies_command() -> int:
    '''CLI functionality for analyzing dependencies.'''
    app_ctx = app_context.app_ctx.get()

    collection_root: str = app_ctx.extra['collection_root']

    errors = check_collection_dependencies(collection_root)

    for error in errors:
        print(error)

    return 3 if errors else 0



# coding: utf-8
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Build command for creating a new major Ansible version."""

import asyncio
import os

import aiohttp
import asyncio_pool
import semantic_version as semver

from antsibull_core import app_context
from antsibull_core.ansible_core import AnsibleCorePyPiClient
from antsibull_core.dependency_files import BuildFile, parse_pieces_file
from antsibull_core.galaxy import GalaxyClient

from .changelog import ChangelogData


def display_exception(loop, context):  # pylint:disable=unused-argument
    print(context.get('exception'))


async def get_version_info(collections, pypi_server_url):
    """
    Return the versions of all the collections and ansible-core
    """
    loop = asyncio.get_running_loop()
    loop.set_exception_handler(display_exception)

    requestors = {}
    async with aiohttp.ClientSession() as aio_session:
        lib_ctx = app_context.lib_ctx.get()
        async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
            pypi_client = AnsibleCorePyPiClient(aio_session, pypi_server_url=pypi_server_url)
            requestors['_ansible_core'] = await pool.spawn(pypi_client.get_versions())
            galaxy_client = GalaxyClient(aio_session)

            for collection in collections:
                requestors[collection] = await pool.spawn(
                    galaxy_client.get_versions(collection))

            collection_versions = {}
            responses = await asyncio.gather(*requestors.values())

    for idx, collection_name in enumerate(requestors):
        collection_versions[collection_name] = responses[idx]

    return collection_versions


# pylint:disable-next=unused-argument
def version_is_compatible(ansible_core_version, collection: str, version: semver.Version,
                          allow_prereleases: bool = False):
    # Metadata for this is not currently implemented.  So everything is rated as compatible
    # as long as it is no prerelease
    if version.prerelease and not allow_prereleases:
        return False
    return True


def find_latest_compatible(ansible_core_version, raw_dependency_versions,
                           allow_prereleases: bool = False):
    # Note: ansible-core compatibility is not currently implemented.  It will be a piece of
    # collection metadata that is present in the collection but may not be present in galaxy.
    # We'll have to figure that out once the pieces are finalized

    # Order versions
    reduced_versions = {}
    for dep, versions in raw_dependency_versions.items():
        # Order the versions
        versions = [semver.Version(v) for v in versions]
        versions.sort(reverse=True)

        # Step through the versions to select the latest one which is compatible
        for version in versions:
            if version_is_compatible(ansible_core_version, dep, version,
                                     allow_prereleases=allow_prereleases):
                reduced_versions[dep] = version
                break

    return reduced_versions


def new_ansible_command():
    app_ctx = app_context.app_ctx.get()
    collections = parse_pieces_file(
        os.path.join(app_ctx.extra['data_dir'], app_ctx.extra['pieces_file']))
    dependencies = asyncio.run(get_version_info(collections, app_ctx.pypi_url))

    ansible_core_version = dependencies.pop('_ansible_core')[0]
    dependencies = find_latest_compatible(
        ansible_core_version, dependencies, allow_prereleases=app_ctx.extra['allow_prereleases'])

    build_filename = os.path.join(app_ctx.extra['dest_data_dir'], app_ctx.extra['build_file'])
    build_file = BuildFile(build_filename)
    build_file.write(app_ctx.extra['ansible_version'], ansible_core_version, dependencies)

    changelog = ChangelogData.ansible(app_ctx.extra['dest_data_dir'])
    changelog.changes.save()

    return 0



# coding: utf-8
# Author: Felix Fontein <tkuratom@redhat.com>
# Author: Toshio Kuratomi <tkuratom@redhat.com>
# License: GPLv3+
# Copyright: Ansible Project, 2020
"""Changelog handling and processing code."""

import asyncio
from collections import defaultdict
import datetime
import glob
import os
import os.path
import tarfile
import tempfile
import typing as t

import aiohttp
import asyncio_pool
import yaml
from packaging.version import Version as PypiVer
from semantic_version import Version as SemVer

from antsibull_changelog.config import PathsConfig, CollectionDetails, ChangelogConfig
from antsibull_changelog.changes import add_release, ChangesData
from antsibull_changelog.changelog_generator import ChangelogGenerator
from antsibull_changelog.utils import collect_versions

from antsibull_core import app_context
from antsibull_core.ansible_core import get_ansible_core
from antsibull_core.dependency_files import DepsFile, DependencyFileData
from antsibull_core.galaxy import CollectionDownloader


class ChangelogData:
    '''
    Data for a single changelog (for a collection, for ansible-core, for Ansible)
    '''

    paths: PathsConfig
    config: ChangelogConfig
    changes: ChangesData
    generator: ChangelogGenerator
    generator_flatmap: bool

    def __init__(self, paths: PathsConfig, config: ChangelogConfig,
                 changes: ChangesData, flatmap: bool = False):
        self.paths = paths
        self.config = config
        self.changes = changes
        self.generator_flatmap = flatmap
        self.generator = ChangelogGenerator(
            self.config, self.changes, plugins=None, fragments=None, flatmap=flatmap)

    @classmethod
    def collection(cls, collection_name: str, version: str,
                   changelog_data: t.Optional[t.Any] = None) -> 'ChangelogData':
        paths = PathsConfig.force_collection('')
        collection_details = CollectionDetails(paths)
        collection_details.namespace, collection_details.name = collection_name.split('.', 1)
        collection_details.version = version
        collection_details.flatmap = False  # TODO!
        config = ChangelogConfig.default(paths, collection_details)
        return cls(paths,
                   config,
                   ChangesData(config, '', changelog_data),
                   flatmap=True)  # TODO!

    @classmethod
    def ansible_core(cls, changelog_data: t.Optional[t.Any] = None) -> 'ChangelogData':
        paths = PathsConfig.force_ansible('')
        collection_details = CollectionDetails(paths)
        config = ChangelogConfig.default(paths, collection_details)
        return cls(paths, config, ChangesData(config, '', changelog_data), flatmap=False)

    @classmethod
    def ansible(cls, directory: t.Optional[str],
                output_directory: t.Optional[str] = None) -> 'ChangelogData':
        paths = PathsConfig.force_ansible('')

        config = ChangelogConfig.default(paths, CollectionDetails(paths), 'Ansible')
        # TODO: adjust the following lines once Ansible switches to semantic versioning
        config.use_semantic_versioning = False
        config.release_tag_re = r'''(v(?:[\d.ab\-]|rc)+)'''
        config.pre_release_tag_re = r'''(?P<pre_release>(?:[ab]|rc)+\d*)$'''

        changelog_path = ''
        if directory is not None:
            changelog_path = os.path.join(directory, 'changelog.yaml')
        changes = ChangesData(config, changelog_path)
        if output_directory is not None:
            changes.path = os.path.join(output_directory, 'changelog.yaml')
        return cls(paths, config, changes, flatmap=True)

    @classmethod
    def concatenate(cls, changelogs: t.List['ChangelogData']) -> 'ChangelogData':
        return cls(
            changelogs[0].paths,
            changelogs[0].config,
            ChangesData.concatenate([changelog.changes for changelog in changelogs]),
            flatmap=changelogs[0].generator_flatmap)

    def add_ansible_release(self, version: str, date: datetime.date, release_summary: str) -> None:
        add_release(self.config, self.changes, [], [], version,
                    codename=None, date=date, update_existing=True,
                    show_release_summary_warning=False)
        release_date = self.changes.releases[version]
        if 'changes' not in release_date:
            release_date['changes'] = {}
        release_date['changes']['release_summary'] = release_summary


def read_file(tarball_path: str, matcher: t.Callable[[str], bool]) -> t.Optional[bytes]:
    with tarfile.open(tarball_path, "r:gz") as tar:
        for file in tar:
            if matcher(file.name):
                file_p = tar.extractfile(file)
                if file_p:
                    with file_p:
                        return file_p.read()
    return None


def read_changelog_file(tarball_path: str, is_ansible_core=False) -> t.Optional[bytes]:
    def matcher(filename: str) -> bool:
        if is_ansible_core:
            return filename.endswith('changelogs/changelog.yaml')
        return filename in ('changelogs/changelog.yaml', 'changelog.yaml')

    return read_file(tarball_path, matcher)


def get_porting_guide_filename(version: PypiVer):
    if version.major == 2 and version.minor == 10:
        basename = 'porting_guide_base'
    else:
        basename = 'porting_guide_core'
    return f"docs/docsite/rst/porting_guides/{basename}_{version.major}.{version.minor}.rst"


def read_porting_guide_file(tarball_path: str, version: PypiVer) -> t.Optional[bytes]:
    filename = get_porting_guide_filename(version)
    return read_file(tarball_path, lambda fn: fn == filename)


class CollectionChangelogCollector:
    collection: str
    versions: t.List[SemVer]
    earliest: SemVer
    latest: SemVer

    changelog: t.Optional[ChangelogData]

    def __init__(self, collection: str, versions: t.ValuesView[str]):
        self.collection = collection
        self.versions = sorted(SemVer(version) for version in versions)
        self.earliest = self.versions[0]
        self.latest = self.versions[-1]
        self.changelog = None

    async def _get_changelog(self, version: SemVer,
                             collection_downloader: CollectionDownloader
                             ) -> t.Optional[ChangelogData]:
        path = await collection_downloader.download(self.collection, version)
        changelog = read_changelog_file(path)
        if changelog is None:
            return None
        changelog_data = yaml.load(changelog, Loader=yaml.SafeLoader)
        return ChangelogData.collection(self.collection, str(version), changelog_data)

    async def _download_changelog_stream(self, start_version: SemVer,
                                         collection_downloader: CollectionDownloader
                                         ) -> t.Optional[ChangelogData]:
        changelog = await self._get_changelog(start_version, collection_downloader)
        if changelog is None:
            return None

        changelog.changes.prune_versions(versions_after=None, versions_until=str(start_version))
        changelogs = [changelog]
        ancestor = changelog.changes.ancestor
        while ancestor is not None:
            ancestor_ver = SemVer(ancestor)
            if ancestor_ver < self.earliest:
                break
            changelog = await self._get_changelog(ancestor_ver, collection_downloader)
            if changelog is None:
                break
            changelog.changes.prune_versions(versions_after=None, versions_until=ancestor)
            changelogs.append(changelog)
            ancestor = changelog.changes.ancestor

        return ChangelogData.concatenate(changelogs)

    async def download(self, collection_downloader: CollectionDownloader):
        missing_versions = set(self.versions)

        while missing_versions:
            missing_version = max(missing_versions)

            # Try to get hold of changelog for this version
            changelog = await self._download_changelog_stream(
                missing_version, collection_downloader)
            if changelog:
                current_changelog = self.changelog
                if current_changelog is None:
                    # If we didn't have a changelog so far, start with it
                    self.changelog = changelog
                    missing_versions -= {SemVer(version) for version in changelog.changes.releases}
                else:
                    # Insert entries from changelog into combined changelog that are missing there
                    for version, entry in changelog.changes.releases.items():
                        sem_version = SemVer(version)
                        if sem_version in missing_versions:
                            current_changelog.changes.releases[version] = entry
                            missing_versions.remove(sem_version)

            # Make sure that this version isn't checked again
            missing_versions -= {missing_version}


class AnsibleCoreChangelogCollector:
    versions: t.List[PypiVer]
    earliest: PypiVer
    latest: PypiVer

    changelog: t.Optional[ChangelogData]

    porting_guide: t.Optional[bytes]

    def __init__(self, versions: t.ValuesView[str]):
        self.versions = sorted(PypiVer(version) for version in versions)
        self.earliest = self.versions[0]
        self.latest = self.versions[-1]
        self.changelog = None
        self.porting_guide = None

    @staticmethod
    async def _get_changelog_file(version: PypiVer,
                                  core_downloader: t.Callable[[str], t.Awaitable[str]]
                                  ) -> t.Optional[ChangelogData]:
        path = await core_downloader(str(version))
        if os.path.isdir(path):
            changelog: t.Optional[ChangelogData] = None
            for root, dummy, files in os.walk(path):
                if 'changelog.yaml' in files:
                    with open(os.path.join(root, 'changelog.yaml'), 'rb') as f:
                        changelog = f.read()
                    changelog_data = yaml.load(changelog, Loader=yaml.SafeLoader)
                    changelog = ChangelogData.ansible_core(changelog_data)
            return changelog
        if os.path.isfile(path) and path.endswith('.tar.gz'):
            changelog = read_changelog_file(path, is_ansible_core=True)
            if changelog is None:
                return None
            changelog_data = yaml.load(changelog, Loader=yaml.SafeLoader)
            return ChangelogData.ansible_core(changelog_data)
        return None

    async def download_changelog(self, core_downloader: t.Callable[[str], t.Awaitable[str]]):
        changelog = await self._get_changelog_file(self.latest, core_downloader)
        if changelog is None:
            return

        changelog.changes.prune_versions(versions_after=None, versions_until=str(self.latest))

        changelogs = [changelog]
        ancestor = changelog.changes.ancestor
        while ancestor is not None:
            ancestor_ver = PypiVer(ancestor)
            if ancestor_ver < self.earliest:
                break
            changelog = await self._get_changelog_file(ancestor_ver, core_downloader)
            if changelog is None:
                break
            changelog.changes.prune_versions(versions_after=None, versions_until=ancestor)
            changelogs.append(changelog)
            ancestor = changelog.changes.ancestor

        self.changelog = ChangelogData.concatenate(changelogs)

    async def download_porting_guide(self, aio_session: 'aiohttp.client.ClientSession'):
        branch_url = 'https://raw.githubusercontent.com/ansible/ansible/devel'

        query_url = f"{branch_url}/{get_porting_guide_filename(self.latest)}"
        async with aio_session.get(query_url) as response:
            self.porting_guide = await response.read()


async def collect_changelogs(collectors: t.List[CollectionChangelogCollector],
                             core_collector: AnsibleCoreChangelogCollector,
                             collection_cache: t.Optional[str]):
    lib_ctx = app_context.lib_ctx.get()
    with tempfile.TemporaryDirectory() as tmp_dir:
        async with aiohttp.ClientSession() as aio_session:
            async with asyncio_pool.AioPool(size=lib_ctx.thread_max) as pool:
                downloader = CollectionDownloader(aio_session, tmp_dir,
                                                  collection_cache=collection_cache)

                async def core_downloader(version):
                    return await get_ansible_core(aio_session, version, tmp_dir)

                requestors = [
                    await pool.spawn(collector.download(downloader)) for collector in collectors
                ]
                requestors.append(
                    await pool.spawn(core_collector.download_changelog(core_downloader)))
                requestors.append(
                    await pool.spawn(core_collector.download_porting_guide(aio_session)))
                await asyncio.gather(*requestors)


class ChangelogEntry:
    version: PypiVer
    version_str: str
    is_ancestor: bool

    prev_version: t.Optional[PypiVer]
    core_versions: t.Dict[PypiVer, str]
    versions_per_collection: t.Dict[str, t.Dict[PypiVer, str]]

    core_collector: AnsibleCoreChangelogCollector
    ansible_changelog: ChangelogData
    collectors: t.List[CollectionChangelogCollector]

    ansible_core_version: str
    prev_ansible_core_version: t.Optional[str]

    removed_collections: t.List[t.Tuple[CollectionChangelogCollector, str]]
    added_collections: t.List[t.Tuple[CollectionChangelogCollector, str]]
    unchanged_collections: t.List[t.Tuple[CollectionChangelogCollector, str]]
    changed_collections: t.List[t.Tuple[CollectionChangelogCollector, str, t.Optional[str], bool]]

    def __init__(self, version: PypiVer, version_str: str,
                 prev_version: t.Optional[PypiVer],
                 ancestor_version: t.Optional[PypiVer],
                 core_versions: t.Dict[PypiVer, str],
                 versions_per_collection: t.Dict[str, t.Dict[PypiVer, str]],
                 core_collector: AnsibleCoreChangelogCollector,
                 ansible_changelog: ChangelogData,
                 collectors: t.List[CollectionChangelogCollector]):
        self.version = version
        self.version_str = version_str
        self.is_ancestor = False if ancestor_version is None else ancestor_version == version
        self.prev_version = prev_version = prev_version or ancestor_version
        self.core_versions = core_versions
        self.versions_per_collection = versions_per_collection
        self.core_collector = core_collector
        self.ansible_changelog = ansible_changelog
        self.collectors = collectors

        self.ansible_core_version = core_versions[version]
        self.prev_ansible_core_version = core_versions.get(prev_version) if prev_version else None

        self.removed_collections = []
        self.added_collections = []
        self.unchanged_collections = []
        self.changed_collections = []
        for collector in collectors:
            if version not in versions_per_collection[collector.collection]:
                if prev_version and prev_version in versions_per_collection[collector.collection]:
                    self.removed_collections.append((
                        collector, versions_per_collection[collector.collection][prev_version]))

                continue

            collection_version: str = versions_per_collection[collector.collection][version]

            prev_collection_version: t.Optional[str] = (
                versions_per_collection[collector.collection].get(prev_version)
                if prev_version else None
            )
            added = False
            if prev_version:
                if not prev_collection_version:
                    self.added_collections.append((collector, collection_version))
                    added = True
                elif prev_collection_version == collection_version:
                    self.unchanged_collections.append((collector, collection_version))
                    continue

            self.changed_collections.append((
                collector, collection_version, prev_collection_version, added))


class CollectionMetadata:
    '''
    Stores metadata about one collection.
    '''

    changelog_url: t.Optional[str]

    def __init__(self, source: t.Optional[t.Mapping[str, t.Any]] = None):
        if source is None:
            source = {}
        self.changelog_url = source.get('changelog-url')


class CollectionsMetadata:
    '''
    Stores metadata about a set of collections.
    '''

    data: t.Dict[str, CollectionMetadata]

    def __init__(self, deps_dir: t.Optional[str]):
        self.data = {}
        if deps_dir is not None:
            collection_meta_path = os.path.join(deps_dir, 'collection-meta.yaml')
            if os.path.exists(collection_meta_path):
                with open(collection_meta_path, 'rb') as collection_meta_file:
                    data = yaml.load(collection_meta_file, Loader=yaml.SafeLoader)
                if data and 'collections' in data:
                    for collection_name, collection_data in data['collections'].items():
                        self.data[collection_name] = CollectionMetadata(collection_data)

    def get_meta(self, collection_name: str) -> CollectionMetadata:
        result = self.data.get(collection_name)
        if result is None:
            result = CollectionMetadata()
            self.data[collection_name] = result
        return result


class Changelog:
    ansible_version: PypiVer
    ansible_ancestor_version: t.Optional[PypiVer]
    entries: t.List[ChangelogEntry]
    core_collector: AnsibleCoreChangelogCollector
    ansible_changelog: ChangelogData
    collection_collectors: t.List[CollectionChangelogCollector]
    collection_metadata: CollectionsMetadata

    def __init__(self,
                 ansible_version: PypiVer,
                 ansible_ancestor_version: t.Optional[PypiVer],
                 entries: t.List[ChangelogEntry],
                 core_collector: AnsibleCoreChangelogCollector,
                 ansible_changelog: ChangelogData,
                 collection_collectors: t.List[CollectionChangelogCollector],
                 collection_metadata: CollectionsMetadata):
        self.ansible_version = ansible_version
        self.ansible_ancestor_version = ansible_ancestor_version
        self.entries = entries
        self.core_collector = core_collector
        self.ansible_changelog = ansible_changelog
        self.collection_collectors = collection_collectors
        self.collection_metadata = collection_metadata


def get_changelog(
        ansible_version: PypiVer,
        deps_dir: t.Optional[str],
        deps_data: t.Optional[t.List[DependencyFileData]] = None,
        collection_cache: t.Optional[str] = None,
        ansible_changelog: t.Optional[ChangelogData] = None
        ) -> Changelog:
    dependencies: t.Dict[str, DependencyFileData] = {}

    ansible_changelog = ansible_changelog or ChangelogData.ansible(directory=deps_dir)
    ansible_ancestor_version_str = ansible_changelog.changes.ancestor
    ansible_ancestor_version = (
        PypiVer(ansible_ancestor_version_str) if ansible_ancestor_version_str else None
    )

    collection_metadata = CollectionsMetadata(deps_dir)

    if deps_dir is not None:
        for path in glob.glob(os.path.join(deps_dir, '*.deps'), recursive=False):
            deps_file = DepsFile(path)
            deps = deps_file.parse()
            version = PypiVer(deps.ansible_version)
            if version > ansible_version:
                print(f"Ignoring {path}, since {deps.ansible_version}"
                      f" is newer than {ansible_version}")
                continue
            dependencies[deps.ansible_version] = deps
    if deps_data:
        for deps in deps_data:
            dependencies[deps.ansible_version] = deps

    core_versions: t.Dict[PypiVer, str] = {}
    versions: t.Dict[str, t.Tuple[PypiVer, DependencyFileData]] = {}
    versions_per_collection: t.Dict[str, t.Dict[PypiVer, str]] = defaultdict(dict)
    for deps in dependencies.values():
        version = PypiVer(deps.ansible_version)
        versions[deps.ansible_version] = (version, deps)
        core_versions[version] = deps.ansible_core_version
        for collection_name, collection_version in deps.deps.items():
            versions_per_collection[collection_name][version] = collection_version

    core_collector = AnsibleCoreChangelogCollector(core_versions.values())
    collectors = [
        CollectionChangelogCollector(collection, versions_per_collection[collection].values())
        for collection in sorted(versions_per_collection.keys())
    ]
    asyncio.run(collect_changelogs(collectors, core_collector, collection_cache))

    changelog = []

    sorted_versions = collect_versions(versions, ansible_changelog.config)
    for index, (version_str, dummy) in enumerate(sorted_versions):
        version, deps = versions[version_str]
        prev_version = None
        if index + 1 < len(sorted_versions):
            prev_version = versions[sorted_versions[index + 1][0]][0]

        changelog.append(ChangelogEntry(
            version,
            version_str,
            prev_version,
            ansible_ancestor_version,
            core_versions,
            versions_per_collection,
            core_collector,
            ansible_changelog,
            collectors))

    return Changelog(
        ansible_version,
        ansible_ancestor_version,
        changelog,
        core_collector,
        ansible_changelog,
        collectors,
        collection_metadata)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['antsibull', 'antsibull.cli', 'antsibull.data', 'antsibull.utils', 'tests']

package_data = \
{'': ['*'], 'antsibull.data': ['debian/*']}

install_requires = \
['antsibull-changelog>=0.14.0',
 'antsibull-core>=1.0.0,<2.0.0',
 'antsibull-docs>=1.0.0,<2.0.0',
 'asyncio-pool',
 'jinja2']

entry_points = \
{'console_scripts': ['antsibull-build = antsibull.cli.antsibull_build:main',
                     'antsibull-lint = antsibull.cli.antsibull_lint:main']}

setup_kwargs = {
    'name': 'antsibull',
    'version': '0.47.0',
    'description': 'Tools for building the Ansible Distribution',
    'long_description': '# antsibull -- Ansible Build Scripts\n[![Python linting badge](https://github.com/ansible-community/antsibull/workflows/Python%20linting/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull/actions?query=workflow%3A%22Python+linting%22+branch%3Amain)\n[![Python testing badge](https://github.com/ansible-community/antsibull/workflows/Python%20testing/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull/actions?query=workflow%3A%22Python+testing%22+branch%3Amain)\n[![dumb PyPI on GH pages badge](https://github.com/ansible-community/antsibull/workflows/👷%20dumb%20PyPI%20on%20GH%20pages/badge.svg?event=push&branch=main)](https://github.com/ansible-community/antsibull/actions?query=workflow%3A%22👷+dumb+PyPI+on+GH+pages%22+branch%3Amain)\n[![Codecov badge](https://img.shields.io/codecov/c/github/ansible-community/antsibull)](https://codecov.io/gh/ansible-community/antsibull)\n\nTooling for building various things related to Ansible\n\nScripts that are here:\n\n* antsibull-build - Builds Ansible-2.10+ from component collections ([docs](docs/build-ansible.rst))\n* antsibull-lint - Soon to be deprecated; collection docs linting functionality is now part of antsibull-docs, and ``changelogs/changelog.yaml`` validation functionality is now part of antsibull-changelog.\n\nThis also includes a [Sphinx extension](https://www.sphinx-doc.org/en/master/) `sphinx_antsibull_ext` which provides a minimal CSS file to render the output of `antsibull-docs` correctly.\n\nRelated projects are [antsibull-changelog](https://pypi.org/project/antsibull-changelog/) and [antsibull-docs](https://pypi.org/project/antsibull-docs/), which are in their own repositories ([antsibull-changelog repository](https://github.com/ansible-community/antsibull-changelog/), [antsibull-docs repository](https://github.com/ansible-community/antsibull-docs/)). Currently both are dependencies of antsibull. Therefore, the scripts contained in them will be available as well when installing antsibull.\n\nYou can find a list of changes in [the Antsibull changelog](./CHANGELOG.rst).\n\nUnless otherwise noted in the code, it is licensed under the terms of the GNU\nGeneral Public License v3 or, at your option, later.\n\nantsibull is covered by the [Ansible Code of Conduct](https://docs.ansible.com/ansible/latest/community/code_of_conduct.html).\n\n## Versioning and compatibility\n\nFrom version 0.1.0 on, antsibull sticks to semantic versioning and aims at providing no backwards compatibility breaking changes **to the command line API (antsibull and antsibull-lint)** during a major release cycle. We might make exceptions from this in case of security fixes for vulnerabilities that are severe enough.\n\nWe explicitly exclude code compatibility. **antsibull is not supposed to be used as a library.** The only exception are potential dependencies with other antsibull projects (currently, none). If you want to use a certain part of antsibull-docs as a library, please create an issue so we can discuss whether we add a stable interface for **parts** of the Python code. We do not promise that this will actually happen though.\n\n## Running from source\n\nPlease note that to run antsibull from source, you need to install some related projects adjacent to the antsibull checkout.  More precisely, assuming you checked out the antsibull repository in a directory `./antsibull/`, you need to check out the following projects in the following locations:\n\n- [antsibull-changelog](https://github.com/ansible-community/antsibull-changelog/) needs to be checked out in `./antsibull-changelog/`.\n- [antsibull-core](https://github.com/ansible-community/antsibull-core/) needs to be checked out in `./antsibull-core/`.\n- [antsibull-docs](https://github.com/ansible-community/antsibull-docs/) needs to be checked out in `./antsibull-docs/`.\n\nThis can be done as follows:\n\n    git clone https://github.com/ansible-community/antsibull-changelog.git\n    git clone https://github.com/ansible-community/antsibull-core.git\n    git clone https://github.com/ansible-community/antsibull-docs.git\n    git clone https://github.com/ansible-community/antsibull.git\n    cd antsibull\n\nScripts are created by poetry at build time.  So if you want to run from a checkout, you\'ll have to run them under poetry::\n\n    python3 -m pip install poetry\n    poetry install  # Installs dependencies into a virtualenv\n    poetry run antsibull-build --help\n\nNote: When installing a package published by poetry, it is best to use pip >= 19.0.  Installing with pip-18.1 and below could create scripts which use pkg_resources which can slow down startup time (in some environments by quite a large amount).\n\n## Creating a new release:\n\nIf you want to create a new release::\n\n    vim pyproject.toml  # Make sure the correct version number is used\n    vim changelogs/fragment/$VERSION_NUMBER.yml  # create \'release_summary:\' fragment\n    antsibull-changelog release --version $VERSION_NUMBER\n    git add CHANGELOG.rst changelogs\n    git commit -m "Release $VERSION_NUMBER."\n    poetry build\n    poetry publish  # Uploads to pypi.  Be sure you really want to do this\n\n    git tag $VERSION_NUMBER\n    git push --tags\n    vim pyproject.toml  # Bump the version number to X.Y.Z.post0\n    git commit -m \'Update the version number for the next release\' pyproject.toml\n    git push\n',
    'author': 'Toshio Kuratomi',
    'author_email': 'a.badger@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ansible-community/antsibull',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
